import com.cycling74.max.*;
import java.io.File;
import java.io.IOException;

import javax.sound.midi.MidiSystem;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.Sequence;
import javax.sound.midi.Track;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.SysexMessage;
import javax.sound.midi.Receiver;
import java.util.LinkedList;



public class midiparty extends MaxObject
{
    private static final MidiTrack[] EMPTYTRACKARRAY = new MidiTrack[0];
    private static final String[] OUTLET_ASSIST = new String[]{"Midi Messages [trackno channel msg byte1 byte2]",
							       "Sysex Messages [trackno sysexdata]",
							       "Meta Messages [trackno metadata]"};
    private File midifile             = null;
    private Sequence sequence         = null;
    private Track[] tracks;
    private float fDivisionType;
    private int i,c;
    private MidiTrack[] miditracks    = null;
    private double ppq_scale_factor = 1;

    /* Attributes  ..........................................................................*/
    /* output raw midievents or text ones */
   private  boolean raw = false;

    /*PPQ to scale the midi file to. These days a lot of sequencers are 384 or greater PPQ
      resolution To accomplish that sort of behavior we would need to run at signal rate and 
      have the the scehduler in audio interrupt. With 96 PPQ we should be able to get up to 
      600+ BPM with scheduler performance set to 1 ms and 300+ BPM with the default scheduler
      setting of 2 ms*/
    private long resolution   = 96;
    /* length of sequence in ticks */
    private long ticklength = 0;
    /* length of sequence in millis */
    private long mslength = 0;    
    /*tickesperbeat or tickeperframe */
    private String  resolutiontype = null;
    /*internal resolution of midifile in resolutiontype */
    private int     internalresolution  = 0;
    /*PPQ or various SMPTE schemes */
    private String  divisiontype = null;

    public midiparty()
    {
	this(null,96);
    }

    public midiparty(String filename)
    {
	this(filename,96);
    }

    public midiparty(String filename,int resolution_)
    {
	init_atom_table();
	miditracks = EMPTYTRACKARRAY;
	if(filename != null)
		open(filename);

	declareIO(1,3);//ShortMessage,Sysex,Meta Message outputs
	declareAttribute("raw");
	resolution = resolution_;
	declareReadOnlyAttribute("internalresolution");
	declareReadOnlyAttribute("ticklength","getticklength");
	declareReadOnlyAttribute("mslength");
	declareReadOnlyAttribute("resolutiontype");
	declareReadOnlyAttribute("divisiontype");
	declareReadOnlyAttribute("numtracks","getnumtracks");
	setOutletAssist(OUTLET_ASSIST);
    }

    private Atom[] getticklength()
    {
	return new Atom[]{Atom.newAtom(ppq_scale_factor * ticklength)};
    }
    /* not sure if this works -tml it don't work. we should not have this a settable attribute. it should
     just be in the constructor*/
    public void resolution(int resolution_)
    {
	if(sequence != null)
	    {
		resolution = resolution_;
		loadtracks(true);
    	    }
    }
    public void dumpinfo()
    {
	int idx = getInfoIdx();
	outlet(idx,new Atom[]{Atom.newAtom("resolution"),Atom.newAtom(resolution)});
	outlet(idx,new Atom[]{Atom.newAtom("internalresolution"),Atom.newAtom(internalresolution)});	
	outlet(idx,new Atom[]{Atom.newAtom("ticklength"),Atom.newAtom(ppq_scale_factor * ticklength)});
	outlet(idx,new Atom[]{Atom.newAtom("mslength"),Atom.newAtom(mslength)});
	outlet(idx,new Atom[]{Atom.newAtom("resolutiontype"),Atom.newAtom(resolutiontype)});
	outlet(idx,new Atom[]{Atom.newAtom("divisiontype"),Atom.newAtom(divisiontype)});
	outlet(idx,new Atom[]{Atom.newAtom("numtracks"),Atom.newAtom(tracks.length)});
	
    }

    public void getresolution()
    {
	outlet(getInfoIdx(),new Atom[]{Atom.newAtom("resolution"),Atom.newAtom(resolution)});
    }

    public Atom[] getnumtracks()
    {
	return new Atom[]{Atom.newAtom(tracks.length)};
    }
    
    public void open()
    {
	String filename = MaxSystem.openDialog();
	if(filename != null)
	    open(filename);
    }

    public void open(String filename)
    {
	String f = MaxSystem.locateFile(filename);
	if(f != null)
		midifile = new File(f);
	else if((midifile = new File(filename)) != null && midifile.exists())
		midifile = midifile;//duh
	else
	    {
	        error("Could not find midi file "+filename);
		return;
	    }
	try {
		sequence = MidiSystem.getSequence(midifile);
	    }
	catch (InvalidMidiDataException e)
	    {

		error("midi file "+midifile.getAbsolutePath()+" is apparently invalid");
		e.printStackTrace();
		return;
	    }
	catch (IOException e)
	    {
		error("error reading midi file "+midifile.getAbsolutePath());
		e.printStackTrace();
		return;
	    }
	if (sequence == null)
	    {
		 error("Cannot retrieve Sequence from."+midifile.getAbsolutePath());
		 return;
	    }
	else
	    {
		ticklength = sequence.getTickLength() ;
		mslength   = sequence.getMicrosecondLength();
		fDivisionType = sequence.getDivisionType();
		internalresolution = sequence.getResolution();
		divisiontype = null;

		if (fDivisionType == Sequence.PPQ)
			divisiontype = "PPQ";
		else if (fDivisionType == Sequence.SMPTE_24)
			divisiontype = "SMPTE_24";
		else if (fDivisionType == Sequence.SMPTE_25)
			divisiontype = "SMPTE_25";
		else if (fDivisionType == Sequence.SMPTE_30DROP)
			divisiontype = "SMPTE_30DROP";
		else if (fDivisionType == Sequence.SMPTE_30)
			divisiontype = "SMPTE_30";

		if (sequence.getDivisionType() == Sequence.PPQ)
			resolutiontype = "ticksperbeat";
		else
			resolutiontype = "ticksperframe";

		tracks = sequence.getTracks();
		//build sequence in memory
		miditracks = new MidiTrack[tracks.length];
		loadtracks(false);
	    }
    }
    /* load track asychronously */
    private void loadtracks(final boolean inplace)
    {
	//calculate scale factor. ppq is our internal PPQ
	ppq_scale_factor = (double)resolution / internalresolution;
	Thread t = new Thread(){
		public void run()
		{
		    for(i = 0; i < tracks.length;i++)
			{
			    // post("loading track "+i);
				   
			    MidiTrack tmp = new MidiTrack(i);
			    int lastinsert = 0;
			    MidiEvent evt;
			    for(c = 0;c < tracks[i].size();c++)
				{
				    evt = tracks[i].get(c);
				    evt.setTick((long) (evt.getTick()*ppq_scale_factor)); 
				    lastinsert = tmp.push_midi_event(lastinsert,evt);
				}
			    if(inplace)
				{
				    //try and figure this out at some point!! we need to stay in the same place somehow
				    //this is hard since it is asynchronous
				    tmp.nexteventindex =(int)( miditracks[i].nexteventindex - 1);
				    tmp.currenttick =(int) (miditracks[i].currenttick * ppq_scale_factor);
				}


			    miditracks[i] = tmp;
			}
		    try{
		    	Thread.sleep(1000);//this is because we need to wait until the patcher is complete to bang out
			//1 sec is sort of conservative
		      	    }catch(InterruptedException e)
		    	{}
		     outlet(getInfoIdx(),new Atom[]{Atom.newAtom("loaded"),Atom.newAtom(midifile.getAbsolutePath())});
		     }

		    };
		    t.start();
    }    
    //run full garbage collect. For instrumentation purposes
    public void gc()
    {
	System.gc();
    }
    public void bang()
    {
	int i;
	for(i = 0;i < miditracks.length;i++)
	    miditracks[i].tick();
    }

    public void reset()
    {
	for(i = 0;i < miditracks.length;i++)
	    {
		miditracks[i].setCurrentTick(0);
		miditracks[i].setNextEventIndex(0);
		miditracks[i].setEnded(false);

	    }
    }

    ////inner class MidiTrack/////
    class MidiTrack
    {
	int track_no;
	int nexteventindex;
	int currenttick;
	LinkedList eventlist;
	boolean ended = false;

	MidiTrack (int track_no_)
	{
	    eventlist = new LinkedList();
	    track_no = track_no_;
	    nexteventindex = 0;
	    currenttick = 0;
	    ended = false;
	}
	
	void setEnded(boolean b)
	{
	    ended = b;
	}
	void setCurrentTick(int ct)
	{
	    currenttick = ct;
	}
	void setNextEventIndex(int nei)
	{
	    nexteventindex = nei;
	}

	//returns index where element was inserted
	int push_midi_event( int start, MidiEvent e)
	{
	    int i;
	    MidiEvent elem = null;
	    int size = eventlist.size();
	    if (e == null)
		return 0;
	
	    if(size == 0)//first element
		{
		    eventlist.addFirst(e);
		    return 0;
		}
	    
	    //System.out.println("size is "+size);
	    for(i = start; i < size;i++)
		{
		    //	System.out.println(" i is "+i);
		    elem = (MidiEvent)eventlist.get(i);
		  
		    if(e.getTick() <= elem.getTick())
			{
			    eventlist.add(i,e);     
			    return i;
			}
		    else if(e.getTick() == elem.getTick())
			{
			    while((elem = (MidiEvent)eventlist.get(i)).getTick() == e.getTick())
				{
				    i++;
				    if(i + 1 >= size - 1)
					{
					    eventlist.addLast(e);//addlast
					    return size - 1;
					}
				}
			    eventlist.add(i+ 1,e);     
			    return i+ 1;//we found where to insert it		   	
			}
	 
		}
	    eventlist.add(size - 1,e);//its tick time is greater than anything previosly inserted	    
	    return size - 1;
	}


	void tick()
	{
	    if(ended)
		return;
	    long netick = ((MidiEvent)eventlist.get(nexteventindex)).getTick();
	    if(currenttick == netick)
		{
		    //		    post("event at tick "+netick);
		    // post("% "+internalresolution+ "is "+netick % internalresolution);
		//we have an event at this tick
		output_events();
		}	   
	    currenttick++;	
	}
	
	void output_events()
	{
	    MidiEvent e;
	    while(true)
		{
		    e = (MidiEvent)eventlist.get(nexteventindex);
		    if(e == null)//end of the list
			break;		
		    else if(currenttick == e.getTick())
			{
			    decode(e.getMessage());
			    nexteventindex++;
			    if(nexteventindex >= eventlist.size())
				{//no more events in this track
				    ended = true;
				    break;
				}
			}		
		    else
			break;
		}
	}

  //decode and output event
  private  void decode(MidiMessage message)
    {
	Atom[] msg = null;
	if (message instanceof ShortMessage)
	    {
		msg = decodeMessage((ShortMessage) message);
		if(msg != null)
		    outlet(0,msg);
	    }
	else if (message instanceof SysexMessage)
	    {
		msg = decodeMessage((SysexMessage) message);
		if(msg != null)
		    outlet(1,msg);
	    }
	else if (message instanceof MetaMessage)
	    {
		msg = decodeMessage((MetaMessage) message);
		if(msg != null)
		    outlet(2,msg);
	    }
    }

	//we create one atom array for use in short messages per instance
    Atom[] sm = new Atom[5];
   private  Atom[] decodeMessage(ShortMessage message)
    {

	switch (message.getCommand())
	    {
	    case 0x80:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0x80 | message.getChannel()]:NOTEOFF;
		sm[3] = atomtable[message.getData1()];//note number
		sm[4] = atomtable[message.getData2()];//velocity
		return sm;
	    case 0x90:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0x90 | message.getChannel()]:NOTEON;
		sm[3] = atomtable[message.getData1()];//note number
		sm[4] = atomtable[message.getData2()];//velocity
		return sm;
	    case 0xa0:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0xa0 | message.getChannel()]:POLYKEYPRESSURE;//polyphonic key pressure
		sm[3] = atomtable[message.getData1()];//note number
		sm[4] = atomtable[message.getData2()];//pressure
		return sm;
	    case 0xb0:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0xb0 | message.getChannel()]:CC;//control change
		sm[3] = atomtable[message.getData1()];//cc#
		sm[4] = atomtable[message.getData2()];//val
		return sm;
	    case 0xc0:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0xc0 | message.getChannel()]:PROGRAM;//program change
		sm[3] = atomtable[message.getData1()];//program #
		return Atom.removeLast(sm);
	    case 0xd0:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0xd0 | message.getChannel()]:KEYPRESSURE;//key pressure
		sm[3] = atomtable[message.getData1()];//note number
		sm[4] = atomtable[message.getData2()];//pressure
		return sm;
	    case 0xe0:
		sm[0] = atomtable[track_no];
		sm[1] = atomtable[message.getChannel()];
		sm[2] = raw ? atomtable[0xe0 | message.getChannel()]:PITCHBEND;//key pressure
		sm[3] = Atom.newAtom(get14bitValue(message.getData1(), message.getData2()));//pitch bend val
		return Atom.removeLast(sm);
	    default:
		return null;
	    }
    }

    Atom[] sysex = new Atom[2];
    private Atom[] decodeMessage(SysexMessage message)
    {
	sysex[0] = atomtable[track_no];
	sysex[1] = raw ? atomtable[0xf7]:SYSEX;//sysex
	return Atom.union(sysex,Atom.newAtom(message.getData()));
    }
	
    private  Atom[] decodeMessage(MetaMessage message)
    {
	byte[]  abMessage = message.getMessage();
	byte[]  abData = message.getData();
	int     nDataLength = message.getLength();
	switch (message.getType())
	    {

	    case 0:
		int     nSequenceNumber = ((abData[0] & 0xFF) << 8) | (abData[1] & 0xFF);
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("sequencenumber"),Atom.newAtom(((abData[0] & 0xFF) << 8) | (abData[1] & 0xFF))};
	    case 1:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("textevent"),Atom.newAtom(new String(abData))};
	    case 2:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("copyright"),Atom.newAtom(new String(abData))};
	    case 3:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("trackname"),Atom.newAtom(new String(abData))};
	    case 4:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("instrname"),Atom.newAtom(new String(abData))};
	    case 5:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("lyrics"),Atom.newAtom(new String(abData))};
	    case 6:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("marker"),Atom.newAtom(new String(abData))};
	    case 7:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("cuepoint"),Atom.newAtom(new String(abData))};
	    case 0x20:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("midichannelprefix"),Atom.newAtom(abData[0] & 0xFF)};
	    case 0x2F:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("endoftrack")};
	    case 0x51:
		int     nTempo = ((abData[0] & 0xFF) << 16)
		    | ((abData[1] & 0xFF) << 8)
		    | (abData[2] & 0xFF);           // tempo in microseconds per beat
		float bpm = convertTempo(nTempo);
		// truncate it to 2 digits after dot
		bpm = (float) (Math.round(bpm*100.0f)/100.0f);
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("bpm"),Atom.newAtom(bpm)};
	    case 0x58:
		return new Atom[]{Atom.newAtom(track_no),Atom.newAtom("timesignature"),Atom.newAtom(new String((abData[0] & 0xFF) + "/" + (1 << (abData[1] & 0xFF))))};
	    default:
		return null;

	    }
    }


    private  int get14bitValue(int nLowerPart, int nHigherPart)
        {
	    return (nLowerPart & 0x7F) | ((nHigherPart & 0x7F) << 7);
        }

    private  float convertTempo(float value) {
	if (value <= 0) {
	    value = 0.1f;
	}
	return 60000000.0f / value;
    }

    }//end class MidiTrack


    /* primative Atom pooling */
    private static Atom[] atomtable = new Atom[255];
    private static boolean atom_table_filled = false;
    private static Atom NOTEON          = Atom.newAtom("noteon");
    private static Atom NOTEOFF         = Atom.newAtom("noteoff");
    private static Atom POLYKEYPRESSURE = Atom.newAtom("polykeypressure");
    private static Atom CC              = Atom.newAtom("cc");
    private static Atom PROGRAM         = Atom.newAtom("program");
    private static Atom KEYPRESSURE     = Atom.newAtom("keypressure");
    private static Atom PITCHBEND       = Atom.newAtom("pitchbend");
    private static Atom SYSEX           = Atom.newAtom("sysex");

    private static void init_atom_table()
    {
	if(!atom_table_filled)
	    {
		for(int i = 0; i < atomtable.length;i++)
		    atomtable[i] = Atom.newAtom(i);
	    }
    }
}






